# Licensed under a 3-clause BSD style license, see LICENSE.
"""
Submodule for helpers to the Dataset-like classes
=================================================

User-facing classes that structure the provenance information, i.e. the history of operations performed on the dataset.

**Available classes:**

* ``Provenance`` (abstract base class).
* ``Origin`` (abstract base class).
* ``ObjectOrigin``.
* ``FileOrigin``.
* ``Transformation``.
* ``Formatting``.
"""

# -----------------------------------------------------------------------------
# Import statements
# -----------------------------------------------------------------------------
from __future__ import absolute_import

import json

from .exceptions import SkhepTypeError
from .py23 import *


# -----------------------------------------------------------------------------
# Provenance-like classes
# -----------------------------------------------------------------------------
class Provenance(object):
    """
    Abstract base class for all classes containing provenance information.

    Trying to instantiate it raises an exception. Instantiate one of its subclasses instead.
    """

    def __init__(self):
        raise SkhepTypeError('Provenance')

    @property
    def detail(self):
        """
        String providing detailed information about the origin, transformation, or formatting.
        """
        raise NotImplementedError

class Origin(Provenance):
    """
    Abstract base class for all classes describing the first object in a provenance list.

    Trying to instantiate it raises an exception. Instantiate one of its subclasses instead.
    """

    def __init__(self):
        raise SkhepTypeError('Origin')


class ObjectOrigin(Origin):
    """
    Declares that the dataset came from some Python object.
    Its history prior to that is unknown.

    Parameters
    ----------
    detail: str
        String providing detailed information about the object origin.

    Examples
    --------
    >>> from skhep.utils import ObjectOrigin
    >>> from array import array
    >>> data = array('i',[1,2,3])
    >>> provenance1 = ObjectOrigin(repr(data))
    >>> provenance1
    <ObjectOrigin>
    >>> provenance1.detail
    "array('i', [1, 2, 3])"

    >>> provenance2 = ObjectOrigin('array_of_ints')
    >>> provenance2.detail
    'array_of_ints'
    """

    def __init__(self, detail):
        if not isinstance(detail, string_types):
            assert False, 'Argument is not of string type!'
        self._detail = detail

    @property
    def detail(self):
        return self._detail
    
    def __repr__(self):
        return "<ObjectOrigin>"

class FileOrigin(Origin):
    """
    Declares that the dataset came from a file or a set of files.

    Parameters
    ----------
    files: str or iterable of str or file objects
        File name(s) or object(s).

    Examples
    --------
    >>> from skhep.utils import FileOrigin
    >>> prov = FileOrigin(['file1.root', 'file2.root','file3.root'])
    >>> prov
    <FileOrigin (3 files)>
    """

    def __init__(self, files):
        if isinstance(files, string_types):
            self.files = (files,)
        else:
            self.files = []
            for x in files:
                if isinstance(x, file):
                    self.files.append(x.name)
                elif isinstance(x, string_types):
                    self.files.append(x)
                else:
                    assert False, ('Argument must be a string filename, an iterable of file objects, '
                                   'or an iterable of string filenames!')
            self.files = tuple(self.files)

    @property
    def detail(self):
        return ",".join(json.dumps(x) for x in self.files)

    def __repr__(self):
        return "<FileOrigin ({0} file{1})>".format(len(self.files),'s' if len(self.files)>1 else '')
        
class Transformation(Provenance):
    """
    Declares that the dataset was transformed by some mathematical operation.

    Parameters
    ----------
    name: str
        String detailing how the dataset got transformed.
    args: iterable, optional
        Optional set of arguments given extra detail on the transformation.

    Examples
    --------
    >>> from skhep.utils import Transformation
    >>> transf = Transformation('all elms * 2')
    >>> transf
    <Transformation(all elms * 2)>
    """

    def __init__(self, name, *args):
        self.name     = name
        self.args     = args

    @property
    def detail(self):
        detail = self.name
        if len(self.args) > 0:
            subdetail = ""
            for x in self.args:
                if isinstance(x, Provenance):
                    subdetail += ", {0}".format(x.detail)
                else:
                    subdetail += ", {0}".format(x)
                if x == self.args[0]:
                    subdetail = subdetail.replace(", ","")
            detail += " ({0})".format(subdetail)
        return detail
            
    def __repr__(self):
        return "<Transformation({0})>".format(self.name)

class Formatting(Provenance):
    """
    Declares that the dataset was reformatted, keeping its semantic meaning, but changed in representation.
    """

    def __init__(self, format, *args):
        self.format = format
        self.args = args

    @property
    def detail(self):
        detail = self.format
        if len(self.args) > 0:
            subdetail = ""
            for x in self.args:
                if isinstance(x, Provenance):
                    subdetail += ", {0}".format(x.detail)
                else:
                    subdetail += ", {0}".format(x)
                if x == self.args[0]:
                    subdetail = subdetail.replace(", ","")
            detail += "({0})".format(subdetail)
        return detail
    
    def __repr__(self):
        return "<Formatting to {0}>".format(self.detail)
      
   
class MultiProvenance(object):
    """
    Class for collecting history of provenances.
    """

    def __init__(self, *args):
        if not all(isinstance(arg, (Provenance, MultiProvenance )) for arg in args):
            raise ValueError("Inputs must be Provenance types!")
        if len(args) == 1 and isinstance(args[0], MultiProvenance):
            args = args[0]
        self._provenances = list(args)
          
    def copy(self):
        return MultiProvenance(*self._provenances)
        
    def __getitem__(self, i):
        return self._provenances[i]
        
    def __repr__(self):
        if len(self._provenances) == 1:
            return repr(self._provenances[0])
        else:
            rep = ""
            for i,provenance in enumerate(self._provenances):
                rep += "{0}: {1}".format(i, provenance)
                rep += "\n" if provenance != self._provenances[-1] else ""
            return rep
            
    @property
    def detail(self):
        if len(self._provenances) == 1:
            return getattr( self._provenances[0], "detail", "")
        else:
            rep = ""
            for i,provenance in enumerate(self._provenances):
                if "\n" in provenance.detail:
                    details = provenance.detail.split("\n")
                    for j, d in enumerate(details):
                        if j == 0:
                            rep += "{0}: {1}\n".format(i, d)
                        else:
                            rep += "   {0}".format(d)
                            rep += "\n" if d != details[-1] else ""
                else:
                    rep += "{0}: {1}".format(i, provenance.detail)
                    
                rep += "\n" if provenance != self._provenances[-1] else ""
                    
            return rep
        
    def __iadd__(self, object):
        if not isinstance( object, ( Provenance , MultiProvenance )):
            raise ValueError("Cannot add a {0} to MultiProvenance!".format(type(object)))
        elif isinstance( object, MultiProvenance):
            self._provenances += object._provenances
        else:
            self._provenances += [object]
        return self
            
    def __add__(self, object):
        multiprov = self.copy()
        multiprov += object
        return multiprov


# Licensed under a 3-clause BSD style license, see LICENSE.
"""
Bayesian Block implementation
=============================

Dynamic programming algorithm for finding the optimal adaptive-width histogram.

* Based on Scargle et al 2012 [1]_
* Initial Python Implementation [2]_
* AstroML Implementation [3]_
* Initial Examination in HEP context [4]_

References
----------
.. [1] http://adsabs.harvard.edu/abs/2012arXiv1207.5578S
.. [2] http://jakevdp.github.com/blog/2012/09/12/dynamic-programming-in-python/
.. [3] https://github.com/astroML/astroML/blob/master/astroML/density_estimation/bayesian_blocks.py
.. [4] https://arxiv.org/abs/1708.00810
"""
from __future__ import absolute_import
from __future__ import division
from skhep.utils.py23 import *
import numpy as np
import pandas as pd


class Prior(object):
    """Helper class for calculating the prior on the fitness function."""
    def __init__(self, p0=0.05, gamma=None):
        self.p0 = p0
        self.gamma = gamma

    def calc(self, N):
        if self.gamma is not None:
            return -np.log(self.gamma)
        else:
            # eq. 21 from Scargle 2012
            return 4 - np.log(73.53 * self.p0 * (N ** -0.478))


def bayesian_blocks(data, weights=None, p0=0.05, gamma=None):
    """Bayesian Blocks Implementation.

    This is a flexible implementation of the Bayesian Blocks algorithm described in Scargle 2012
    [1]_.  It has been modified to natively accept weighted events, for ease of use in HEP
    applications.

    Args:
        data (array):
            Input data values (one dimensional, length N). Repeat values are allowed.

        weights (array_like, optional):
            Weights for data (otherwise assume all data points have a weight of 1).  Must be same
            length as data.

            Defaults to None.

        p0 (float, optional):
            False-positive rate, between 0 and 1.  A lower number places a stricter penalty against
            creating more bin edges, thus reducing the potential for false-positive bin edges.  In
            general, the larger the number of bins, the small the p0 should be to prevent the
            creation of spurious, jagged bins.

            Defaults to 0.05.

        gamma (float, optional):
            If specified, then use this gamma to compute the general prior form, p ~ gamma^N. If
            gamma is specified, p0 is ignored.

            Defaults to None.


    Returns:
        edges (ndarray):
            Array containing the (N+1) bin edges

    Examples:
        Unweighted data:

        >>> d = np.random.normal(size=100)
        >>> bins = bayesian_blocks(d, p0=0.01)

        Unweighted data with repeats:

        >>> d = np.random.normal(size=100)
        >>> d[80:] = d[:20]
        >>> bins = bayesian_blocks(d, p0=0.01)

        Weighted data:

        >>> d = np.random.normal(size=100)
        >>> w = np.random.uniform(1,2, size=100)
        >>> bins = bayesian_blocks(d, w, p0=0.01)


    See Also:
        skhep.visual.MplPlotter.hist:
            Histogram plotting function which can natively make use of bayesian blocks.
    """
    # validate input data
    data = np.asarray(data, dtype=float)
    assert data.ndim == 1

    # validate input weights
    if weights is not None:
        weights = np.asarray(weights)
    else:
        # set them to 1 if not given
        weights = np.ones_like(data)

    # initialize the prior
    prior = Prior(p0, gamma)

    # Place data and weights into a DataFrame.
    # We want to sort the data array (without losing the associated weights), and combine duplicate
    # data points by summing their weights together.  We can accomplish all this with `groupby`

    df = pd.DataFrame({'data': data, 'weights': weights})
    gb = df.groupby('data').sum()
    data = gb.index.values
    weights = gb.weights.values

    N = weights.size

    # create length-(N + 1) array of cell edges
    edges = np.concatenate([data[:1],
                            0.5 * (data[1:] + data[:-1]),
                            data[-1:]])
    block_length = data[-1] - edges

    # arrays to store the best configuration
    best = np.zeros(N, dtype=float)
    last = np.zeros(N, dtype=int)

    # -----------------------------------------------------------------
    # Start with first data cell; add one cell at each iteration
    # -----------------------------------------------------------------
    # last = core_loop(N, block_length, weights, fitfunc, best, last)
    for R in range(N):
        # Compute fit_vec : fitness of putative last block (end at R)

        # T_k: width/duration of each block
        T_k = block_length[:R + 1] - block_length[R + 1]

        # N_k: number of elements in each block
        N_k = np.cumsum(weights[:R + 1][::-1])[::-1]

        # evaluate fitness function
        fit_vec = N_k * (np.log(N_k/T_k))

        # penalize function with prior
        A_R = fit_vec - prior.calc(R + 1)
        A_R[1:] += best[:R]

        i_max = np.argmax(A_R)
        last[R] = i_max
        best[R] = A_R[i_max]

    # -----------------------------------------------------------------
    # Now find changepoints by iteratively peeling off the last block
    # -----------------------------------------------------------------
    change_points = np.zeros(N, dtype=int)
    i_cp = N
    ind = N
    while True:
        i_cp -= 1
        change_points[i_cp] = ind
        if ind == 0:
            break
        ind = last[ind - 1]
    change_points = change_points[i_cp:]

    return edges[change_points]



# Licensed under a 3-clause BSD style license, see LICENSE.
"""
Submodule for helpers to deal with dependencies
===============================================
"""

# Two ideas:
#
# Strict softimport tries to load at startup (like a normal Python import)
# but hides the fact that it failed until the user tries to use it.
#
# Lazy softimport doesn't even try to load until the user tries to use it.
#
# Both have the same semantics: the user only needs to have installed the
# dependencies he or she will need. But strict puts all the load-time at
# startup while lazy puts all the load-time at first use.
#
# We should have a policy and stick to it. Strict is more typical of Python
# modules, but lazy would spare the user the 1.09 second lag on startup
# if they have ROOT on their system but don't intend to use it in this session.
#
# **Edit:** Another advantage of lazy softimport: if the user tries to use the
# module, it fails with an ImportError, then they install it and try to use
# it again in the same Python session, this will work with lazy softimport
# and not with strict softimport. For this reason, I'm changing the default
# to lazy=True and would advocate that we _only_ do lazy softimports.


class DelayedImportError(object):
    # When a module is strictly softimported but is not available, this
    # object is a placeholder for the module. The user will get the
    # original error when they try to use it.

    def __init__(self, err):
        self.err = err

    def __getattr__(self, attr):
        raise self.err


class LazyModule(object):
    # When a module is lazily softimported, it will always be represented
    # with this object as a placeholder. It will repeatedly attempt to
    # import the module every time the user tries to use it (unlike strict,
    # which gives up after the first attempt on startup).
    #
    # Unlike strict, the module will always be wrapped by this object,
    # even if the module is available. Users will be able to see
    # the sausage-making if they're inclined to do `type(root_numpy)`.

    def __init__(self, name):
        self.name = name
        self.module = None

    def __getattr__(self, attr):
        if self.module is None:
            self.module = __import__(self.name)
        return getattr(self.module, attr)


def softimport(modulename, lazy=True):
    """
    The function that one calls to import a module softly.
    """
    # Once we vote on "always strict" or "always lazy," we should _take out_
    # the other functionality so that it isn't even an option.

    if lazy:
        return LazyModule(modulename)
    else:
        try:
            return __import__(modulename)
        except ImportError as err:
            return DelayedImportError(err)



# Licensed under a 3-clause BSD style license, see LICENSE.
"""
Vector classes
==============

Two vector classes are available:

* ``Vector3D``     : a 3-dimensional vector.
* ``LorentzVector``: a Lorentz vector, i.e. a 4-dimensional Minkowski space-time vector
                     or a 4-momentum vector.
                     The metric is (-1,-1,-1,+1).
"""

# -----------------------------------------------------------------------------
# Import statements
# -----------------------------------------------------------------------------
from __future__ import absolute_import

from skhep.utils.py23 import *

from skhep.utils.exceptions import *

from math import sqrt, atan2, cos, sin, acos, degrees, log, pi, sinh

# -----------------------------------------------------------------------------
# Vector class in 3D
# -----------------------------------------------------------------------------
class Vector3D(object):
    """
    Vector class in 3 dimensions.

    Constructors:
        __init__(x=0., y=0., z=0.)
        origin()
        frompoint(x, y, z)
        fromvector(avector)
        fromsphericalcoords(r, theta, phi)
        fromcylindricalcoords(rho, phi, z)
        fromiterable(values)
    """

    def __init__(self, x=0., y=0., z=0.):
        """Default constructor.

        Examples
        --------
        >>> v1 = Vector3D()
        >>> v1
        <Vector3D (x=0.0,y=0.0,z=0.0)>
        >>> v2 = Vector3D(1,2,3)
        >>> v2
        <Vector3D (x=1,y=2,z=3)>
        """
        self.__values = [x, y, z]

    @classmethod
    def origin(cls):
        """Shortcut constuctor for the origin (x=0.,y=0.,z=0.).
        Equivalent to the default constructor Vector3D().
        """
        return cls(0., 0., 0.)

    @classmethod
    def frompoint(cls, x, y, z):
        """Constructor from an explicit space point."""
        return cls(x, y, z)

    @classmethod
    def fromvector(cls, other):
        """Copy constructor."""
        return cls(other.x, other.y, other.z)

    @classmethod
    def fromsphericalcoords(cls, r, theta, phi):
        """Constructor from a space point specified in spherical coordinates.

        Parameters
        ----------
        r     : radius, the radial distance from the origin (r > 0)
        theta : inclination in radians (theta in [0, pi] rad)
        phi   : azimuthal angle in radians (phi in [0, 2pi) rad)
        """
        x = r * sin(theta) * cos(phi)
        y = r * sin(theta) * sin(phi)
        z = r * cos(theta)
        return cls(x, y, z)

    @classmethod
    def fromcylindricalcoords(cls, rho, phi, z):
        """Constructor from a space point specified in cylindrical coordinates.

        Parameters
        ----------
        rho : radial distance from the z-axis (rho > 0)
        phi : azimuthal angle in radians (phi in [-pi, pi) rad)
        z   : height
        """
        x = rho * cos(phi)
        y = rho * sin(phi)
        z = z
        return cls(x, y, z)

    @classmethod
    def fromiterable(cls, values):
        """Constructor from a suitable iterable object.
        Suitable means here that all entries are numbers
        and the length equals 3.
        """
        values = list(values)
        if not len(values)==3:
            raise ValueError( 'Input iterable length = {0}! Please check your inputs.'.format(len(values)) )
        for i, v in enumerate(values):
            if not isinstance( v, (int,float) ):
                raise ValueError( 'Component #{0} is not a number!'.format(i) )
        return cls(values[0], values[1], values[2])

    @property
    def x(self):
        """Return the x, aka first coordinate at position 0."""
        return self.__values[0]

    @x.setter
    def x(self, value):
        """Sets x, aka first coordinate at position 0."""
        self.__values[0] = value

    @property
    def y(self):
        """Return the y, aka second coordinate at position 1."""
        return self.__values[1]

    @y.setter
    def y(self, value):
        """Sets y, aka second coordinate at position 1."""
        self.__values[1] = value

    @property
    def z(self):
        """Return the z, aka third coordinate at position 2."""
        return self.__values[2]

    @z.setter
    def z(self, value):
        """Sets z, aka third coordinate at position 2."""
        self.__values[2] = value

    @property
    def rho(self):
        """Return the cylindrical coordinate rho."""
        return sqrt(self.x**2 + self.y**2)

    @property
    def r(self):
        """Return the spherical coordinate r."""
        return self.mag

    def costheta(self):
        """Return the cosine of the spherical coordinate theta."""
        if self.mag == 0.:
            return 1.
        else:
            return self.z / self.mag

    def theta(self, deg=False):
        """Return the spherical coordinate theta.

        Parameters
        ----------
        deg : float, optional
            Return the angle in degrees (default is radians).
        """
        theta = acos(self.costheta())
        return theta if not deg else degrees(theta)

    def phi(self, deg=False):
        """Return the spherical or cylindrical coordinate phi.

        Parameters
        ----------
        deg : float, optional
            Return the angle in degrees (default is radians).
        """
        phi = atan2(self.y, self.x)
        return phi if not deg else degrees(phi)

    def set(self, x, y, z):
        """Update the vector components all at once."""
        self.__values = [x, y, z]

    def __setitem__(self, i, value):
        """Update/set the ith vector component (commencing at 0, of course)."""
        try:
            self.__values[i] = value
        except IndexError:
            raise IndexError(
                'Vector3D is of length {0} only!'.format(len(self)))

    def __getitem__(self, i):
        """Get the ith vector component (commencing at 0, of course)."""
        try:
            return self.__values[i]
        except IndexError:
            raise IndexError( 'Vector3D is of length {0} only!'.format(len(self)))

    def tolist(self):
        """Return the vector as a list."""
        return list(self.__values)

    def __len__(self):
        """Length of the vector, i.e. the number of elements = 3."""
        return len(self.__values)

    @property
    def mag(self):
        """Magnitude, a.k.a. norm, of the vector."""
        return sqrt(self.mag2)

    @property
    def mag2(self):
        """Square of the magnitude, a.k.a. norm, of the vector."""
        return sum(v ** 2 for v in self.__values)

    def __abs__ ( self ) :
        """Get the absolute value for the vector.

        Example
        -------
        >>> v = ...
        >>> a = abs(v)
        """
        return self.mag

    def copy(self) :
        """Get a copy of the vector.

        Example
        -------
        >>> v = ...
        >>> v1 = v.copy()
        """
        return Vector3D( self[0] , self[1] , self[2] )

    def unit(self):
        """Return the normalized vector, i.e. the unit vector along the direction of itself."""
        mag = self.mag
        if mag > 0. and mag != 1. :
            return Vector3D.fromiterable([v / mag for v in self.__values])
        else:
            return self

    def __iadd__(self, other):
        """(self)Addition with another vector, i.e. self+other.

        Example
        -------
        >>> v1  = ...
        >>> v2  = ...
        >>> v1 += v2
        """
        if not isinstance ( other ,  Vector3D ) :
            raise InvalidOperationError("invalid operation '+=' between a 'Vector3D' and a '{0}'".format(other.__class__.__name__))
        self.__values[0] += other.__values[0]
        self.__values[1] += other.__values[1]
        self.__values[2] += other.__values[2]
        return self

    def __isub__(self, other):
        """(self)Subtraction with another vector, i.e. self+other.

        Example
        -------
        >>> v1  = ...
        >>> v2  = ...
        >>> v1 -= v2
        """
        if not isinstance ( other ,  Vector3D ) :
            raise InvalidOperationError("invalid operation '-=' between a 'Vector3D' and a '{0}'".format(other.__class__.__name__))
        self.__values[0] -= other.__values[0]
        self.__values[1] -= other.__values[1]
        self.__values[2] -= other.__values[2]
        return self

    def __add__(self, other):
        """Addition with another vector, i.e. self+other."""
        if not isinstance ( other ,  Vector3D ) : return NotImplemented
        v = self.copy()
        v+= other
        return v

    def __sub__(self, other):
        """Subtraction with another vector, i.e. self-other."""
        if not isinstance ( other ,  Vector3D ) : return NotImplemented
        v = self.copy()
        v-= other
        return v

    def __imul__(self, other):
        """Scaling of the vector by a number.

        Example
        -------
        >>> v = ...
        >>> v *= 2
        """
        if isinstance ( other , ( int , float ) ) :
            return Vector3D.fromiterable ( [v * other for v in self.__values ] )
        else:
            raise InvalidOperationError("invalid operation '*=' between a 'Vector3D' and a '{0}'".format(other.__class__.__name__))

    def __itruediv__(self, number):
        """Scaling of the vector by a number.

        Example
        -------
        >>> v  = ...
        >>> v /= 2
        """
        if not isinstance ( number , ( int , float ) ) :
            raise InvalidOperationError("invalid operation '/=' between a 'Vector3D' and a '{0}'".format(number.__class__.__name__))
        elif 0 == number : raise ZeroDivisionError
        self *= ( 1.0/number )
        return self

    __idiv__ = __itruediv__

    def __mul__(self, other):
        """Multiplication of the vector by either another vector or a number.
        Multiplication of two vectors is equivalent to the dot product, see dot(...).

        Example
        -------
        >>> v2 = v1 * 2
        >>> number = v1 * v3
        """
        if isinstance ( other ,  Vector3D ) :
            return self.dot(other)
        v = self.copy()
        v *= other
        return v

    def __rmul__(self, other):
        """Right multiplication of the vector by either another vector or a number."""
        return self.__mul__(other)

    def __truediv__(self, number):
        """Division of the vector by a number."""
        v = self.copy()
        v /= number
        return v

    __div__ = __truediv__

    def __eq__  ( self , other ) :
        """Equality to another vector, or, equality to zero.

        Example
        -------
        >>> v1 = ...
        >>> v2 = ...
        >>> print v1 == v2
        >>> print v1 == 0
        """
        from skhep.math.numeric import isequal

        ## comparsion with scalar zero, very useful  in practice
        if isinstance ( other , ( float , int , long ) ) and isequal ( other , 0 ) :
            return isequal ( self[0] , 0 ) and isequal ( self[1] , 0 ) and isequal ( self[2] , 0 )
        elif not isinstance ( other , Vector3D) :
            return NotImplemented
        ##
        return isequal ( self[0] , other[0] ) and isequal ( self[1] , other[1] ) and isequal ( self[2] , other[2] )

    def __ne__  (self, other) :
        """Non-equality to another vector.

        Example
        -------
        >>> v1 = ...
        >>> v2 = ...
        >>> print v1 != v2
        """
        return not ( self == other )

    def __nonzero__ ( self ) :
        """Non-zero vector?

        Example
        -------
        >>> vct = ...
        >>> if vct : ...
        """
        return 0 != self.mag2

    __bool__ = __nonzero__

    def __iter__(self):
        """Iterator implementation for the vector components."""
        return self.__values.__iter__()

    def dot(self, other):
        """Dot product with another vector."""
        return sum(v1 * v2 for v1, v2 in zip(self.__values, other.__values))

    def cross(self, other ):
        """Cross product with another vector."""
        return Vector3D( self[1] * other[2] - self[2] * other[1],
                         self[2] * other[0] - self[0] * other[2],
                         self[0] * other[1] - self[1] * other[0]
                        )

    def rotate(self, angle, *args):
        """Rotate vector by a given angle (in radians) around a given axis."""
        if len(args) == 1 and isinstance(args[0], Vector3D):
            ux, uy, uz = args[0].__values
        elif len(args) == 1 and len(args[0]) == 3:
            ux, uy, uz = args[0]
        elif len(args) == 3:
            ux, uy, uz = args
        else:
            raise TypeError('Input object not a Vector3D nor an iterable with 3 elements.')

        for i, u in enumerate((ux, uy, uz)):
            if not isinstance( u, (int,float) ):
                raise ValueError( 'Component #{0} is not a number!'.format(i) )

        norm = sqrt(ux**2 + uy**2 + uz**2)
        if norm != 1.0:
            ux = ux / norm; uy = uy / norm; uz = uz / norm;
        c, s = cos(angle), sin(angle)
        c1 = 1. - c

        xp = (c + ux**2 * c1) * self.x + (ux * uy * c1 - uz * s) * self.y \
            + (ux * uz * c1 + uy * s) * self.z
        yp = (ux * uy * c1 + uz * s) * self.x + (c + uy**2 * c1) * self.y \
            + (uy * uz * c1 - ux * s) * self.z
        zp = (ux * uz * c1 - uy * s) * self.x + (uy * uz * c1 + ux * s) * self.y \
            + (c + uz**2 * c1) * self.z

        return Vector3D(xp, yp, zp)

    def rotatex(self, angle):
        """Rotate vector by a given angle (in radians) around the x axis."""
        return self.rotate(angle, 1, 0, 0)

    def rotatey(self, angle):
        """Rotate vector by a given angle (in radians) around the y axis."""
        return self.rotate(angle, 0, 1, 0)

    def rotatez(self, angle):
        """Rotate vector by a given angle (in radians) around the z axis."""
        return self.rotate(angle, 0, 0, 1)

    def cosdelta ( self , other ) :
        """Get cos(angle) with respect to another vector
        """
        m1 = self.mag2
        if 0 >= m1 : return 1.0
        m2 = other.mag2
        if 0 >= m2 : return 1.0

        r = self.dot( other ) / sqrt ( m1 * m2 )
        return max ( -1.0 , min ( 1.0 , r ) )


    def angle(self, other, deg=False):
        """
        Angle with respect to another vector.

        Parameters
        ----------
        deg : float, optional
            Return the angle in degrees (default is radians).
        """
        cd = self.cosdelta ( other )
        return acos(cd) if not deg else degrees(acos(cd))

    def isparallel(self, other):
        """Check if another vector is parallel.
        Two vectors are parallel if they have the same direction but not necessarily the same magnitude.
        """
        from skhep.math.numeric import isequal
        return isequal ( self.cosdelta(other)  , 1 )

    def isantiparallel(self, other):
        """Check if another vector is antiparallel.
        Two vectors are antiparallel if they have opposite direction but not necessarily the same magnitude.
        """
        from skhep.math.numeric import isequal
        return isequal ( self.cosdelta(other) , -1 )

    def iscollinear ( self , other ) :
        """Check if another vector is collinear
        Two vectors are collinear if they have parallel or antiparallel
        """
        from skhep.math.numeric import isequal
        return isequal ( abs ( self.cosdelta ( other ) ) , 1 )

    def isopposite ( self , other):
        """Two vectors are opposite if they have the same magnitude but opposite direction."""
        from skhep.math.numeric import isequal
        added = self + other
        return added == 0

    def isperpendicular(self, other):
        """Check if another vector is perpendicular."""
        from skhep.math.numeric import isequal
        return isequal ( self.dot ( other ) , 0 , scale = max( self.mag2 , other.mag2 ) )

    def __repr__(self):
        """Class representation."""
        return "<Vector3D (x={0},y={1},z={2})>".format(*self.__values)

    def __str__(self):
        """Simple class representation."""
        return str(tuple(self.__values))

#-----------------------------------------------------------------------------
# Lorentz vector class
#-----------------------------------------------------------------------------
class LorentzVector(object):
    """
    Class representing a Lorentz vector,
    either a 4-dimensional Minkowski space-time vector or a 4-momentum vector.
    The 4-vector components can be seen as (x,y,z,t) or (px,py,pz,E).

    Constructors:
        __init__(x=0., y=0., z=0., t=0.)
        from4vector(avector)
        from3vector(vector3d, t)
    """
    def __init__(self, x=0., y=0., z=0., t=0.):
        """Default constructor.

        Example
        -------
        >>> v1 = LorentzVector()
        >>> v1
        """
        self.__vector3d = Vector3D(x, y, z)
        self.__t = t


    @classmethod
    def from4vector(cls, other):
        """Copy constructor."""
        return cls(other.x, other.y, other.z, other.t)

    @classmethod
    def from3vector(cls, vector3d, t):
        """Constructor from a Vector3D and the time/energy component."""
        return cls(vector3d.x, vector3d.y, vector3d.z, t)

    @classmethod
    def fromiterable(cls, values):
        """Constructor from a suitable iterable object.
        Suitable means here that all entries are numbers
        and the length equals 4.
        """
        values = list(values)
        if not len(values)==4:
            raise ValueError( 'Input iterable length = {0}! Please check your inputs.'.format(len(values)) )
        for i, v in enumerate(values):
            if not isinstance( v, (int,float) ):
                raise ValueError( 'Component #{0} is not a number!'.format(i) )
        return cls(values[0], values[1], values[2], values[3])

    @property
    def x(self):
        """Return the coordinate x, aka first coordinate at position 0."""
        return self.__vector3d.x

    @x.setter
    def x(self, value):
        """Sets x, aka first coordinate at position 0."""
        self.__vector3d.x = value

    @property
    def y(self):
        """Return the coordinate x, aka second coordinate at position 1."""
        return self.__vector3d.y

    @y.setter
    def y(self, value):
        """Sets y, aka second coordinate at position 1."""
        self.__vector3d.y = value

    @property
    def z(self):
        """Return the coordinate z, aka third coordinate at position 2."""
        return self.__vector3d.z

    @z.setter
    def z(self, value):
        """Sets z, aka third coordinate at position 2."""
        self.__vector3d.z = value

    @property
    def vector(self):
        """Return a copy of the vector of spatial components."""
        return self.__vector3d.copy()

    @property
    def t(self):
        """Return the time/energy component, aka coordinate at position 3."""
        return self.__t

    @t.setter
    def t(self, value):
        """Sets t, aka coordinate at position 3."""
        self.__t = value

    def costheta(self):
        """Return the cosinus of the spherical coordinate theta."""
        return self.__vector3d.costheta()

    def theta(self, deg=False):
        """
        Return the spherical coordinate theta.

        Parameters
        ----------
        deg : float, optional
            Return the angle in degrees (default is radians).
        """
        return self.__vector3d.theta(deg)

    def phi(self, deg=False):
        """
        Return the spherical or cylindrical coordinate phi.

        Parameters
        ----------
        deg : float, optional
            Return the angle in degrees (default is radians).
        """
        return self.__vector3d.phi(deg)

    @property
    def px(self):
        """Return the Vector3D coordinate px, aka first momentum coordinate at position 0."""
        return self.__vector3d.x

    @px.setter
    def px(self, value):
        """Sets px, aka first momentum coordinate at position 0."""
        self.__vector3d.x = value

    @property
    def py(self):
        """Return the Vector3D coordinate px, aka second momentum coordinate at position 1."""
        return self.__vector3d.y

    @py.setter
    def py(self, value):
        """Sets py, aka second momentum coordinate at position 1."""
        self.__vector3d.y = value

    @property
    def pz(self):
        """Return the Vector3D coordinate pz, aka third momentum coordinate at position 2."""
        return self.__vector3d.z

    @pz.setter
    def pz(self, value):
        """Sets pz, aka third momentum coordinate at position 2."""
        self.__vector3d.z = value

    @property
    def e(self):
        """Return the energy/time component, aka momentum coordinate at position 3."""
        return self.__t

    @e.setter
    def e(self, value):
        """Sets e, aka momentum coordinate at position 3."""
        self.__t = value

    def set(self, x, y, z, t):
        """Update/set all components at once."""
        self.__vector3d = Vector3D(x, y, z)
        self.__t = t

    def setpxpypzm(self, px, py, pz, m):
        """Set the px,py,pz components and the mass."""
        self.__vector3d.x = px; self.__vector3d.y = py; self.__vector3d.z = pz
        if m > 0.:
            self.__t = sqrt(px**2 + py**2 + pz**2 + m**2)
        else:
            self.__t = sqrt(px**2 + py**2 + pz**2 - m**2)

    def setpxpypze(self, px, py, pz, e):
        """Set the px,py,pz components and the energy."""
        self.set(px,py,pz,e)

    def setptetaphim(self, pt, eta, phi, m):
        """ Set the transverse momentum, the pseudorapidity, the angle phi and the mass."""
        px, py, pz = pt*cos(phi), pt*sin(phi), pt*sinh(eta)
        self.setpxpypzm(px,py,pz,m)

    def setptetaphie(self, pt, eta, phi, e):
        """ Set the transverse momentum, the pseudorapidity, the angle phi and the energy."""
        px, py, pz = pt*cos(phi), pt*sin(phi), pt*sinh(eta)
        self.setpxpypze(px,py,pz,e)

    def tolist(self):
        """Return the LorentzVector as a list."""
        return list(self.__vector3d) + [self.__t]

    def __setitem__(self, i, value):
        """Update/set the ith vector component (commencing at 0, of course)."""
        try:
            if i == 3: self.__t = value
            else: self.__vector3d[i] = value
        except IndexError:
            raise IndexError(
                'LorentzVector is of length {0} only!'.format(len(self)))

    def __getitem__(self, i):
        """Get the ith vector component (commencing at 0, of course)."""
        try:
            return self.tolist()[i]
        except IndexError:
            raise IndexError( 'LorentzVector is of length {0} only!'.format(len(self)))

    def __len__(self):
        """Length of the LorentzVector, i.e. the number of elements = 4."""
        return len(self.tolist())

    @property
    def p(self):
        """Return the momentum, aka norm of the momentum vector."""
        return self.__vector3d.mag

    @property
    def pt(self):
        """Return the transverse momentum, aka transverse component of the momentum vector."""
        return self.perp

    @property
    def et(self):
        """Return the transverse energy."""
        return self.e * ( self.pt / self.p )

    @property
    def m(self):
        """Return the invariant mass."""
        return self.mag

    @property
    def m2(self):
        """Return the square of the invariant mass."""
        return self.mag2

    @property
    def mass(self):
        """Return the invariant mass."""
        return self.mag

    @property
    def mass2(self):
        """Return the square of the invariant mass."""
        return self.mag2

    @property
    def mt(self):
        """Return the transverse mass."""
        return self.transversemass

    @property
    def mt2(self):
        """Return the square of the transverse mass."""
        return self.transversemass2

    @property
    def transversemass(self):
        """Return the transverse mass."""
        mt2 = self.transversemass2
        return sqrt(mt2) if mt2 >= 0. else -sqrt(-mt2)

    @property
    def transversemass2(self):
        """Return the square of the transverse mass."""
        return self.e**2 - self.pz**2

    @property
    def beta(self):
        """Return :math:`\\beta = v/c`."""
        return self.p / self.e

    @property
    def gamma(self):
        """Return :math:`\\gamma = 1/\\sqrt{1-\\beta^2}`."""
        if self.beta < 1:
            return 1. / sqrt(1. - self.beta**2)
        else:
            return 10E10

    @property
    def eta(self):
        """Return the pseudorapidity."""
        if abs(self.costheta()) < 1.:
            return -0.5 * log( (1. - self.costheta())/(1. + self.costheta()) )
        else:
            return 10E10 if self.z > 0 else -10E10

    @property
    def boostvector(self):
        """Return the spatial component divided by the time component."""
        return Vector3D(self.x / self.t, self.y / self.t, self.z / self.t)

    @property
    def pseudorapidity(self):
        """"Return the pseudorapidity. Alternative to eta() method."""
        return self.eta

    @property
    def rapidity(self):
        """Return the rapidity."""
        return 0.5 * log( (self.e + self.pz)/(self.e - self.pz) )

    @property
    def mag(self):
        """Magnitude, a.k.a. norm, of the Lorentz vector."""
        mag2 = self.mag2
        return sqrt(mag2) if mag2 >= 0. else -sqrt(-mag2)

    @property
    def mag2(self):
        """Square of the magnitude, a.k.a. norm, of the Lorentz vector."""
        return self.t**2 - self.__vector3d.mag2

    @property
    def perp2(self):
        """Square of the transverse component, in the (x,y) plane, of the spatial components."""
        return self.x**2 + self.y**2

    @property
    def perp(self):
        """Transverse component of the spatial components."""
        return sqrt(self.perp2)

    def copy(self) :
        """Get a copy of the LorentzVector.

        Example
        -------
        >>> v = ...
        >>> v1 = v.copy()
        """
        return LorentzVector( self[0] , self[1] , self[2] , self[3] )

    def __iadd__(self, other):
        """(self)Addition with another LorentzVector, i.e. self+other.

        Example
        -------
        >>> v1  = ...
        >>> v2  = ...
        >>> v1 += v2
        """
        if not isinstance ( other ,  LorentzVector ) :
            raise InvalidOperationError("invalid operation '+=' between a 'LorentzVector' and a '{0}'".format(other.__class__.__name__))
        self.__vector3d += other.__vector3d
        self.__t += other.__t
        return self

    def __isub__(self, other):
        """(self)Subtraction with another LorentzVector, i.e. self+other.

        Example
        -------
        >>> v1  = ...
        >>> v2  = ...
        >>> v1 -= v2
        """
        if not isinstance ( other ,  LorentzVector ) :
            raise InvalidOperationError("invalid operation '-=' between a 'LorentzVector' and a '{0}'".format(other.__class__.__name__))
        self.__vector3d -= other.__vector3d
        self.__t -= other.__t
        return self

    def __add__(self, other):
        """Addition with another LorentzVector, i.e. self+other."""
        v = self.copy()
        v+= other
        return v

    def __sub__(self, other):
        """Subtraction with another LorentzVector, i.e. self-other."""
        v = self.copy()
        v-= other
        return v


    def __imul__(self, other):
        """Scaling of the Lorentz vector with a number.

        Example
        -------
        >>> v = ...
        >>> v *= 2
        """
        if isinstance ( other , ( int , float ) ) :
            return LorentzVector.fromiterable ( [v * other for v in self.tolist() ] )
        else:
            raise InvalidOperationError("invalid operation '*=' between a 'LorentzVector' and a '{0}'".format(other.__class__.__name__))

    def __itruediv__(self, number):
        """Scaling of the Lorentz vector with a number.

        Example
        -------
        >>> v  = ...
        >>> v /= 2
        """
        if not isinstance ( number , ( int , float ) ) :
            raise InvalidOperationError("invalid operation '/=' between a 'LorentzVector' and a '{0}'".format(number.__class__.__name__))
        elif 0 == number : raise ZeroDivisionError
        self *= ( 1.0/number )
        return self

    __idiv__ = __itruediv__

    def __mul__(self, other):
        """Multiplication of the LorentzVector by either another LorentzVector or a number.
        Multiplication of two LorentzVector is equivalent to the dot product, see dot(...).

        Example
        -------
        >>> v2 = v1 * 2
        >>> number = v1 * v3
        """
        if isinstance ( other , LorentzVector ):
            return self.dot(other)
        v = self.copy()
        v *= other
        return v

    def __rmul__(self, other):
        """Right multiplication of the LorentzVector by either another LorentzVector or a number."""
        return self.__mul__(other)

    def __truediv__(self, number):
        """Division of the LorentzVector by a number."""
        v = self.copy()
        v /= number
        return v

    __div__ = __truediv__

    def __eq__  ( self , other ) :
        """Equality to another LorentzVector, or, equality to zero.

        Example
        -------
        >>> v1 = ...
        >>> v2 = ...
        >>> print v1 == v2
        >>> print v1 == 0
        """
        from skhep.math.numeric import isequal

        ## comparsion with scalar zero, very useful  in practice
        if isinstance ( other , ( float , int , long ) ) and isequal ( other , 0 ) :
            return isequal ( self[0] , 0 ) and isequal ( self[1] , 0 ) and isequal ( self[2] , 0 ) \
             and isequal ( self[3] , 0 )
        elif not isinstance ( other , LorentzVector ) :
            return NotImplemented
        ##
        return isequal ( self[0] , other[0] ) and isequal ( self[1] , other[1] ) and isequal ( self[2] , other[2] ) \
         and isequal ( self[3] , other[3] )

    def __ne__  (self, other) :
        """Non-equality to another Lorentz vector.

        Example
        -------
        >>> v1 = ...
        >>> v2 = ...
        >>> print v1 != v2
        """
        return not ( self == other )

    def __iter__(self):
        """Iterator implementation for the Lorentz vector components."""
        return self.tolist().__iter__()

    def boost(self, *args):
        """Apply a Lorentz boost on the Lorentz vector."""
        if len(args) == 1 and isinstance(args[0], Vector3D):
            bx, by, bz = args[0].x, args[0].y, args[0].z
        elif len(args) == 1 and len(args[0]) == 3:
            bx, by, bz = args[0]
        elif len(args) == 3:
            bx, by, bz = args
        else:
            raise TypeError('Input object not a Vector3D nor an iterable with 3 elements.')

        for i, b in enumerate((bx, by, bz)):
            if not isinstance( b, (int,float) ):
                raise ValueError( 'Component #{0} is not a number!'.format(i) )

        b2 = bx**2 + by**2 + bz**2
        gamma = 1. / sqrt( 1. - b2 )
        bp = bx * self.x + by * self.y + bz * self.z
        if b2 > 0.:
            gamma2 = ( gamma - 1. ) / b2
        else:
            gamma2 = 0.

        xp = self.x + gamma2 * bp * bx - gamma * bx * self.t
        yp = self.y + gamma2 * bp * by - gamma * by * self.t
        zp = self.z + gamma2 * bp * bz - gamma * bz * self.t
        tp = gamma * ( self.t - bp )

        return LorentzVector(xp, yp, zp, tp)

    def rotate(self, angle, *args):
        """Rotate vector by a given angle (in radians) around a given axis."""

        v3p = self.__vector3d.rotate(angle, *args)

        return LorentzVector().from3vector(v3p, self.t)

    def rotatex(self, angle):
        """Rotate vector by a given angle (in radians) around the x axis."""
        return self.rotate(angle, 1, 0, 0)

    def rotatey(self, angle):
        """Rotate vector by a given angle (in radians) around the y axis."""
        return self.rotate(angle, 0, 1, 0)

    def rotatez(self, angle):
        """Rotate vector by a given angle (in radians) around the z axis."""
        return self.rotate(angle, 0, 0, 1)

    def dot(self, other):
        """Dot product with another Lorentz vector."""
        return self.t * other.t - self.__vector3d * other.__vector3d

    def deltaeta(self, other):
        """Return the pseudorapidity difference, :math:`\\Delta \\eta`, with another Lorentz vector."""
        return self.eta - other.eta

    def deltaphi(self, other):
        """Return the phi angle difference, :math:`\\Delta \\phi`, with another Lorentz vector."""
        dphi = self.phi() - other.phi()
        while dphi > pi:
            dphi -= 2*pi
        while dphi < -pi:
            dphi += 2*pi
        return dphi

    def deltar(self, other):
        """Return :math:`\\Delta R` the distance in (eta,phi) space with another Lorentz vector, defined as:
        :math:`\\Delta R = \\sqrt{(\\Delta \\eta)^2 + (\\Delta \\phi)^2}`
        """
        return sqrt( self.deltaeta(other)**2 + self.deltaphi(other)**2 )

    def isspacelike(self):
        """Check if Lorentz Vector is space-like."""
        from skhep.math.numeric import isequal
        return self.mag2 < 0. and not isequal(self.mag2, 0.)

    def istimelike(self):
        """Check if Lorentz Vector is time-like."""
        from skhep.math.numeric import isequal
        return self.mag2 > 0. and not isequal(self.mag2, 0.)

    def islightlike(self):
        """Check if Lorentz Vector is light-like."""
        from skhep.math.numeric import isequal
        return isequal(self.mag2, 0.)

    def __repr__(self):
        """Class representation."""
        return "<LorentzVector (x={0},y={1},z={2},t={3})>".format(*self.tolist())

    def __str__(self):
        """Simple class representation."""
        return str(tuple(self.tolist()))



# Licensed under a 3-clause BSD style license, see LICENSE.
# Brian Pollack, 2017
from __future__ import absolute_import
from __future__ import division
import sys

from skhep.utils.py23 import *

from numbers import Number
from collections import Iterable
import colorsys

import numpy as np

import matplotlib as mpl
from matplotlib import pyplot as plt
from matplotlib import colors, gridspec
import matplotlib.cbook as cbook
from matplotlib.ticker import MaxNLocator
from matplotlib.patches import Polygon

import pandas as pd

from skhep.modeling import bayesian_blocks
from skhep.visual.fill_between_steps import fill_between_steps

import warnings
warnings.filterwarnings('ignore')


class MplPlotter(object):
    def __init__(self, show_backend=False):
        if show_backend:
            import matplotlib
            print(matplotlib.get_backend())

    @staticmethod
    def hist(x, bins='auto', range=None, weights=None, errorbars=False, normed=False, scale=None,
             stacked=False, histtype='stepfilled', **kwargs):
        """Enhanced histogram, based on `hist` function from matplotlib and astroML.
        The main additional features are the ability to use data-driven binning algorithms, the
        addition of errorbars, scaling options (like dividing all bin values by their widths), and
        marker-style draw options.  `hist` function wraps `HistContainer` class, which does the
        majority of work.

        Args:
            x (array, or list of array):
                Array of data to be histogrammed.

            bins (int or List or str, optional): If `int`, `bins` number of equal-width bins are
                generated.  The width is determined by either equal divison of the given `range`, or
                equal division between the first and last data point if no `range` is specified.

                If `List`, bin edges are taken directly from `List` (can be unequal width).

                If `str`, then it must be one of::

                        'blocks' : use bayesian blocks for dynamic bin widths.

                        'auto' : use 'auto' feature from numpy.histogram.

                        'fd', 'doane', 'scott', 'rice', 'sturges', 'sqrt' : see numpy.histogram for
                            details.
                            (https://docs.scipy.org/doc/numpy/reference/generated/numpy.histogram.html)

                Defaults to 'auto'.

            range (tuple or None, optional): If specificed, data will only be considered and shown
                for the range given.  Otherwise, `range` will be between the highest and lowest
                datapoint.

                Defaults to None.

            weights (array or None, optional): Weights associated with each data point.  If
                specified, bin content will be equal to the sum of all relevant weights.

                Defaults to None.

            errorbars (boolean or array, optional): If True, errorbars will be calculated and
                displayed based on the `err_*` arguments.  The errorbars will be appropriately
                modified if `scale` and/or `normed` is True. If an array is specificed, those values
                will be used (and will not be modifed by any other methods).

                Defaults to False.

            normed (boolean, optional): If True, histogram will be normalized such that the integral
                over all bins with be equal to 1.  In general, this will NOT mean that the sum of
                all bin contents will be 1, unless all bin widths are equal to 1. If used with
                `scale` option, normalization happens before scale.

                Defaults to False

            scale (Number or 'binwidth', optional): If Number, all bin contents are multiplied by
                the given value.  If 'binwidth', every bin content is divided by the bin width. If
                used with `normed` option, scaling occurs after normalization ('binwidth' will be
                ignored in this case, because it is handled automatically when normalizing).

                Defaults to None

            stacked (boolean, optional): If True, multiple input data sets will be layered on top of
                each other, such that the height of each bin is the sum of all the relevant dataset
                contributions.  If used with errorbars, the bars will be associated with the bin
                totals, not the individual components.

                Defaults to False.

            histtype (stepfilled', 'step', 'bar', or 'marker'): Draw options for histograms.
                'stepfilled', 'step', and 'bar' inherit from `matplotlib`.  'marker' places a single
                point at the center of each bin (best used with error bars).  'marker' will throw an
                exception if used with 'stacked' option.

                Defaults to 'stepfilled'.

            **kwargs:
                * ax (matplotlib axes instance):
                    Specify the Axes on which to draw the histogram.  If not specified, then the
                    current active axes will be used, or a new axes instance will be generated.

                    Defaults to current axis.

                * err_style ('band' or 'line'):
                    Draw style for errorbars.

                    Defaults depend on `histtype`,
                    where histtype='stepfilled' corresponds to 'band', and all others correspond to
                    'line'.

                * err_color ('auto' or valid `matplotlib` colors):
                    Color for error bars.  If 'auto' is
                    chosen and `stacked` is False, The color will be a slightly darker version of
                    the associated histogram color.  If `stacked` is True, color will be the next in
                    the current color cycle.

                    Defaults to 'auto'.

                * err_type ('sumW2' or 'gaussian'):
                   Method of calculating error bars, if no error is given.  The 'gaussian' method
                   displays the error of each bin as the square root of the bin content.   The
                   'sumW2' method displays the error as the square of the sum of the squares of the
                   weights.

                   Defaults to 'gaussian' if no weights are given, else 'sumW2'.

                * err_return (bool):
                    Return additional objects corresponding to the values of the error associated
                    with each bin, and the patches used to display the error bars.

                    Defaults to 'False'

            Note:
                Other keyword arguments are described in `pylab.hist()`.

        Returns:
            (bin_content, bin_edges, bin_errors (optional), patches)

            bin_content (array or list of array):
                The values of the histogram bins. See normed or
                density and weights for a description of the possible semantics. If input x is an
                array, then this is an array of length nbins. If input is a sequence arrays [data1,
                data2,..], then this is a list of arrays with the values of the histograms for each
                of the arrays in the same order.

            bin_edges (array):
                The edges of the bins. Length nbins + 1 (nbins left edges and right edge of last
                bin). Always a single array even when multiple data sets are passed in.

            bin_errors (array or list of arrays, optional):
                The values of the error bars associated with each bin.  It is only returned if
                `errorbars` is `True` and `err_return` is True.

            bin_patches (list or list of lists):
                Silent list of individual patches used to create the histogram or list of such list
                if multiple input datasets.


        Brian Pollack, 2017
        """

        # Generate a histogram object

        hist_con = HistContainer(x, bins, range, weights, errorbars, normed, scale, stacked,
                                 histtype, **kwargs)

        if hist_con.err_return:
            return hist_con.bin_content, hist_con.bin_edges, hist_con.bin_err, hist_con.vis_object
        else:
            return hist_con.bin_content, hist_con.bin_edges, hist_con.vis_object

    @staticmethod
    def ratio_plot(hist_dict1, hist_dict2, bins=None, range=None, ratio_range=None,
                   err_style='band', err_color='dimgray', ratio_mode='default', grid=False,
                   unity_line='red', logx=False):
        '''Function for creating ratio plots (comparing two histograms by dividing their bin content).
        The call structure is very similar to producing two individual histograms, with additional
        arguments specifying the nature of the ratio plot.  The number of bins and ranges for both
        histograms must be equal.

        Note:
            Addition documentation soon.'''

        bin_range = range
        del range

        bins, bin_range = _check_args_ratio(hist_dict1, hist_dict2, bins, bin_range)
        hist_dict1['bins'] = bins
        hist_dict1['range'] = bin_range

        fig = plt.figure()
        gs = gridspec.GridSpec(2, 1, height_ratios=[3, 1])
        ax1 = fig.add_subplot(gs[0])
        if logx:
            ax1.set_xscale("log", nonposx='clip')
        ax2 = fig.add_subplot(gs[1], sharex=ax1)
        # ax1.grid(True)
        # ax2.grid(True)
        plt.setp(ax1.get_xticklabels(), visible=False)
        fig.subplots_adjust(hspace=0.001)

        hist_dict1['ax'] = ax1
        hist_dict2['ax'] = ax1

        hist_con1 = HistContainer(**hist_dict1)
        bin_edges = hist_con1.bin_edges
        bin_range = (bin_edges[0], bin_edges[-1])
        hist_dict2['bins'] = bin_edges
        hist_dict2['range'] = bin_range
        hist_con2 = HistContainer(**hist_dict2)
        ax1.set_xlim(bin_range)

        if hist_con1.stacked:
            bc1 = hist_con1.bin_content[-1]
        else:
            bc1 = hist_con1.bin_content
        if hist_con2.stacked:
            bc2 = hist_con2.bin_content[-1]
        else:
            bc2 = hist_con2.bin_content

        berr1 = getattr(hist_con1, 'bin_err', np.zeros(len(bc1)))
        berr2 = getattr(hist_con2, 'bin_err', np.zeros(len(bc2)))

        with np.errstate(divide='ignore', invalid='ignore'):
            ratio = bc1/bc2
            ratio_err = ratio*np.sqrt((berr1/bc1)**2+(berr2/bc2)**2)
        ratio_err_hi = ratio + ratio_err
        ratio_err_low = ratio - ratio_err

        ratio[ratio == 0] = np.nan

        fill_between_steps(ax2, hist_con1.bin_edges, ratio_err_hi, ratio_err_low,
                           step_where='pre', linewidth=0, color=err_color, alpha=0.2, zorder=10)

        ax2.errorbar(hist_con1.bin_centers, ratio, yerr=None,
                     xerr=[hist_con1.bin_centers-hist_con1.bin_edges[0:-1],
                           hist_con1.bin_edges[1:]-hist_con1.bin_centers], fmt='d',
                     color=err_color)

        if unity_line:
            ax2.axhline(1, linewidth=3, color=unity_line, zorder=0)
        ax2.yaxis.set_major_locator(MaxNLocator(nbins=4, prune='upper') )
        if ratio_range:
            ax2.set_ylim(ratio_range)
        else:
            ax2.set_ylim((0, 2.0))

        return (ax1, ax2), (hist_con1.bin_content, hist_con1.bin_edges, hist_con1.vis_object), \
            (hist_con2.bin_content, hist_con2.bin_edges, hist_con2.vis_object)


def _check_args_ratio(hist_dict1, hist_dict2, bins, bin_range):

    # check that 'bins' is not conflicting
    if bins is not None:
        _bins = bins
    elif 'bins' in hist_dict1:
        _bins = hist_dict1['bins']
    elif 'bins' in hist_dict2:
        _bins = hist_dict2['bins']
    else:
        _bins = 'auto'

    if (('bins' in hist_dict1 and hist_dict1['bins'] != _bins) or
            ('bins' in hist_dict2 and hist_dict2['bins'] != _bins)):
        raise KeyError('`bins` arg is inconsistent.  All declared `bins` args must be equal')

    # check that 'range' is not conflicting

    if bin_range is not None:
        _bin_range = bin_range
    elif 'range' in hist_dict1:
        _bin_range = hist_dict1['range']
    elif 'range' in hist_dict2:
        _bin_range = hist_dict2['range']
    else:
        _bin_range = None

    if (('range' in hist_dict1 and hist_dict1['range'] != _bin_range) or
            ('range' in hist_dict2 and hist_dict2['range'] != _bin_range)):
        raise KeyError('`range` arg is inconsistent.  All declared `range` args must be equal')

    return _bins, _bin_range


class HistContainer(object):
    """Class to hold histogram properties and members."""
    def __init__(self, x, bins='auto', range=None, weights=None, errorbars=False, normed=False,
                 scale=None, stacked=False, histtype='stepfilled', **kwargs):

        if weights is None:
            self.has_weights = False
        else:
            self.has_weights = True
        x, weights = self._checks_and_wrangling(x, weights)

        # Prevent hiding of builtins, define members from args
        self.bin_range = range
        del range
        self.bins = bins
        if isinstance(errorbars, Iterable):
            self.bin_err = errorbars
            self.errorbars = 'given'
        elif errorbars:
            self.errorbars = 'calc'
        else:
            self.errorbars = False
        self.normed = normed
        self.scale = scale
        self.stacked = stacked
        self.histtype = histtype
        self._arg_init(kwargs, weights)
        self._df_binning_init(x, weights)
        self.do_redraw = False

        if self.normed:
            self.normalize()
        if self.scale:
            self.rescale(self.scale)

        if self.histtype == 'marker':
            self.vis_object = self.ax.plot(self.bin_centers, self.bin_content, **self.hist_dict)

        elif self.do_redraw and not self.histtype == 'marker':
            self.redraw()

        if self.errorbars:
            self.draw_errorbars()
        if self.logy:
            self.set_log_vals()

    def _checks_and_wrangling(self, x, w):
        # Manage the input data in the same fashion as mpl
        if np.isscalar(x):
            x = [x]

        input_empty = (np.size(x) == 0)

        # Massage 'x' for processing.
        if input_empty:
            x = np.array([[]])
        elif mpl.__version__ < '2.1.0':
            x = cbook._reshape_2D(x)
        else:
            x = cbook._reshape_2D(x, 'x')

        self.n_data_sets = len(x)  # number of datasets

        # We need to do to 'weights' what was done to 'x'
        if w is not None:
            if mpl.__version__ < '2.1.0':
                w = cbook._reshape_2D(w)
            else:
                w = cbook._reshape_2D(w, 'w')

        if w is not None and len(w) != self.n_data_sets:
            raise ValueError('weights should have the same shape as x')

        if w is not None:
            for xi, wi in zip(x, w):
                if wi is not None and len(wi) != len(xi):
                    raise ValueError('weights should have the same shape as x')

        return x, w

    def _arg_init(self, kwargs, w):
        # Break up the kwargs into different chunks depending on the arg, and set various defaults.

        # Scaling by `binwidth` is handled by default during normalization
        if self.normed and self.scale == 'binwidth':
            self.scale = None

        if self.histtype == 'marker' and self.stacked:
            raise ValueError('Do not stack with markers, that would be silly')

        if self.histtype == 'marker' and self.n_data_sets > 1:
            raise ValueError('`marker` histtype does not currently support multiple input datasets')

        if self.histtype == 'barstacked' and not self.stacked:
            self.histtype = 'stepfilled'
            self.stacked = True

        # Get current axis
        self.ax = kwargs.pop('ax', plt.gca())

        # Group arguments into different dicts
        # For error bars:
        self.err_dict = {}
        self.err_dict['errorbars'] = self.errorbars

        if self.histtype in ['stepfilled', 'bar']:
            self.err_dict['err_style'] = kwargs.pop('err_style', 'band')
        else:
            self.err_dict['err_style'] = kwargs.pop('err_style', 'line')

        self.err_dict['err_color'] = kwargs.pop('err_color', 'auto')

        self.err_dict['suppress_zero'] = kwargs.pop('suppress_zero', False)

        if self.has_weights:
            self.err_dict['err_type'] = kwargs.pop('err_type', 'sumW2')
        else:
            self.err_dict['err_type'] = kwargs.pop('err_type', 'gaussian')

        self.err_dict['err_x'] = kwargs.pop('err_x', True)
        self.err_dict['err_return'] = kwargs.pop('err_return', False)
        self.err_return = self.err_dict['err_return']
        if self.err_return and not self.errorbars:
            raise KeyError('Cannot set `err_return=True` if `errorbars=False`')

        # tweak histogram styles for `band` err_style

        if self.err_dict['err_style'] == 'band' and self.errorbars and not self.stacked:
            if 'edgecolor' not in kwargs:
                kwargs['edgecolor'] = 'k'
            if 'linewidth' not in kwargs:
                kwargs['linewidth'] = 1

        # For data-driven binning
        self.bin_dict = {}
        if isinstance(self.bins, str):
            if 'gamma' in kwargs:
                self.bin_dict['gamma'] = kwargs.pop('gamma')
            if 'p0' in kwargs:
                self.bin_dict['p0'] = kwargs.pop('p0')

        self.hist_dict = kwargs
        if self.histtype != 'marker':
            self.hist_dict['histtype'] = self.histtype

        # set some marker defaults
        if self.histtype == 'marker':
            if 'marker' not in self.hist_dict:
                self.hist_dict['marker'] = 'o'
            if 'linestyle' not in self.hist_dict:
                self.hist_dict['linestyle'] = ''

        if 'alpha' not in self.hist_dict:
            if self.histtype in ['marker', 'step']:
                self.hist_dict['alpha'] = 1
            else:
                self.hist_dict['alpha'] = 0.5

        if 'linewidth' not in self.hist_dict and self.histtype == 'step':
            self.hist_dict['linewidth'] = 2

        self.logy = self.hist_dict.pop('log', False)

    def _df_binning_init(self, data, weights):
        '''Do an initial binning to get bin edges, total hist range, and break each set of data and
        weights into a dataframe (easier to handle errorbar calculation moving forward)'''

        # If a range is given, mask data using that range to allow smart binning algos to
        # bin data properly.  This will impact a predetermined bin-edge input.

        if self.bin_range is not None:
            for d in range(self.n_data_sets):
                range_mask = (data[d]>=self.bin_range[0])&(data[d]<=self.bin_range[-1])
                data[d] = data[d][range_mask]
            if weights is not None:
                weights[d] = weights[d][range_mask]

        # If bin edges are already determined, than skip initial histogramming
        self.bin_edges = None
        if isinstance(self.bins, Iterable) and not isinstance(self.bins, str):
            self.bin_edges = self.bins
            if self.bin_range is None:
                self.bin_range = (self.bin_edges[0], self.bin_edges[-1])
            else:
                # If range is given, it gets priority over self-defined bins
                range_mask = (self.bin_edges>=self.bin_range[0])&(self.bin_edges<=self.bin_range[-1])
                self.bin_edges = self.bin_edges[range_mask]

        # If bin edges need to be determined, there's a few different cases to consider
        else:
            if self.stacked:
                _n_data_sets = 1
                b_data = [np.concatenate(data)]
                if self.has_weights:
                    b_weights = [np.concatenate(weights)]
                else:
                    b_weights = None
            else:
                _n_data_sets = self.n_data_sets
                b_data = data
                b_weights = weights

            if self.bin_range is None:
                xmin = np.inf
                xmax = -np.inf
                for i in range(_n_data_sets):
                    if len(data[i]) > 0:
                        xmin = min(xmin, min(b_data[i]))
                        xmax = max(xmax, max(b_data[i]))
                self.bin_range = (xmin, xmax)

            # Special case for Bayesian Blocks
            if self.bins in ['block', 'blocks']:

                # Single data-set or stacked
                if _n_data_sets == 1:

                    if self.has_weights:
                        b_weights = b_weights[0]
                    else:
                        b_weights = None
                    self.bin_edges = bayesian_blocks(data=b_data[0], weights=b_weights,
                                                     **self.bin_dict)
                else:
                    raise ValueError('Cannot use Bayesian Blocks with multiple, unstacked datasets')

            else:
                _, self.bin_edges = np.histogram(b_data, bins=self.bins, weights=b_weights,
                                                 range=self.bin_range)

        self.widths = np.diff(self.bin_edges)
        self.bin_centers = self.bin_edges[:-1]+self.widths*0.5

        # Now put the data into dataframes with the weights and bins
        self.df_list = []
        for i in range(self.n_data_sets):
            if weights is None:
                df = pd.DataFrame({'data': data[i]})
            else:
                df = pd.DataFrame({'data': data[i], 'weights': weights[i]})
            df_bins = pd.cut(df.data, self.bin_edges, include_lowest=True)
            df['bins'] = df_bins
            self.df_list.append(df)

        # Make the initial histograms
        if self.histtype == 'marker':
            self.bin_content, _ = np.histogram(data, self.bin_edges, weights=weights,
                                               range=self.bin_range)
        else:
            self.bin_content, _, self.vis_object = self.ax.hist(data, self.bin_edges,
                                                                weights=weights,
                                                                range=self.bin_range,
                                                                stacked=self.stacked,
                                                                **self.hist_dict)

            # if self.stacked and self.errorbars and self.histtype == 'stepfilled':
            #     plt.setp(self.vis_object[-1][0], edgecolor='k')
            #     plt.setp(self.vis_object[-1][0], linewidth=2)

        self.bin_content_orig = self.bin_content[:]

        if self.errorbars == 'calc' and not (self.normed or self.scale):
            self.calc_bin_error(hist_mod='default')

    def normalize(self):
        self.do_redraw = True
        data = [df.data for df in self.df_list]
        if self.has_weights:
            weights = [df.weights for df in self.df_list]
            if self.stacked:
                weights = np.concatenate(weights)
        elif self.n_data_sets == 1 or self.stacked:
            weights = None
        else:
            weights = [None for df in self.df_list]

        if self.n_data_sets == 1:
            self.bin_content, _ = np.histogram(data, self.bin_edges, weights=weights,
                                               range=self.bin_range, density=True)
        elif self.stacked:
            total_bin_content, _ = np.histogram(np.concatenate(data), self.bin_edges,
                                                weights=weights, range=self.bin_range, density=True)
            bin_scales = np.divide(total_bin_content, self.bin_content[-1])
            for i in range(self.n_data_sets):
                self.bin_content[i] = np.multiply(bin_scales, self.bin_content[i])

        else:
            self.bin_content = []
            for i, d in enumerate(data):
                self.bin_content.append(np.histogram(d, self.bin_edges, weights=weights[i],
                                                     range=self.bin_range, density=True)[0])

        if self.errorbars == 'calc':
            self.calc_bin_error(hist_mod='norm')

    def rescale(self, scale):
        self.do_redraw = True

        if isinstance(scale, Number):
            self.bin_content = np.multiply(self.bin_content, scale)
            if self.errorbars == 'calc':
                self.calc_bin_error(hist_mod='scale', scale=scale, exist=self.normed)

        elif scale == 'binwidth':
            widths = np.diff(self.bin_edges)
            self.bin_content = np.divide(self.bin_content, widths)
            if self.errorbars == 'calc':
                self.calc_bin_error(hist_mod='scale', scale=1.0/widths, exist=self.normed)

    def calc_bin_error(self, hist_mod='default', scale=None, exist=False):

        # make new error bars if they haven't been calc'd
        if not exist:
            data = [df.data for df in self.df_list]
            if self.stacked:
                data = np.concatenate(data)

            if self.has_weights:
                weights = [df.weights for df in self.df_list]
                if self.stacked:
                    weights = np.concatenate(weights)
            elif self.n_data_sets == 1 or self.stacked:
                weights = None
            else:
                weights = [None for df in self.df_list]

            if self.n_data_sets == 1 or self.stacked:
                bin_content_no_norm, _ = np.histogram(data, self.bin_edges, weights=weights,
                                                      range=self.bin_range)
            else:
                bin_content_no_norm = []
                for i, d in enumerate(data):
                    bin_content_no_norm.append(np.histogram(d, self.bin_edges, weights=weights[i],
                                                            range=self.bin_range)[0])

            if self.err_dict['err_type'] == 'gaussian':
                bin_err_tmp = np.sqrt(bin_content_no_norm)
                # bin_err_tmp = np.sqrt(self.bin_content)

            elif self.err_dict['err_type'] == 'sumW2':
                bin_err_tmp = []
                if self.stacked:
                    df_list_tmp = [pd.concat(self.df_list, ignore_index=True)]
                else:
                    df_list_tmp = self.df_list

                for df in df_list_tmp:
                    df['weights2'] = np.square(df.weights)
                    bin_err_tmp.append(np.sqrt((
                        df.groupby('bins')['weights2'].sum().fillna(0).values)))

                bin_err_tmp = bin_err_tmp[-1]

            elif self.err_dict['err_type'] == 'poisson':
                err_low = np.asarray([poisson_error(bc, self.err_dict['suppress_zero'])[0] for bc
                                      in bin_content_no_norm])
                err_hi = np.asarray([poisson_error(bc, self.err_dict['suppress_zero'])[1] for bc in
                                     bin_content_no_norm])
                bin_err_tmp = np.asarray([err_low, err_hi])

            else:
                raise KeyError('`err_type: {0}` not implemented'.format(self.err_dict['err_type']))

            # Modifiy the error bars if needed (due to normalization or scaling)
            if hist_mod == 'default':
                self.bin_err = bin_err_tmp

            elif hist_mod == 'norm':
                if self.stacked:
                    bc = self.bin_content[-1]
                else:
                    bc = self.bin_content
                self.bin_err = bin_err_tmp*(np.divide(bc, bin_content_no_norm))

            elif hist_mod == 'scale':
                self.bin_err = np.multiply(bin_err_tmp, scale)

        # if errors already exist due to norm calc
        else:
            self.bin_err = np.multiply(self.bin_err, scale)

    def draw_errorbars(self):
        if self.n_data_sets == 1:
            bin_height = [self.bin_content]
            bin_err = [self.bin_err]
            if self.histtype != 'marker':
                vis_object = [self.vis_object]
            else:
                vis_object = self.vis_object
        elif self.stacked:
            bin_height = [self.bin_content[-1]]
            bin_err = [self.bin_err]
            vis_object = [self.vis_object]
        else:
            bin_height = self.bin_content
            bin_err = self.bin_err
            vis_object = self.vis_object

        if not self.stacked:
            n_data_sets_eff = self.n_data_sets
        else:
            n_data_sets_eff = 1
        for i in range(n_data_sets_eff):
            if self.err_dict['err_color'] == 'auto' and not self.stacked:
                if self.histtype == 'marker':
                    err_color = colors.to_rgba(vis_object[i].get_markerfacecolor(),
                                               vis_object[i]._alpha)
                elif self.histtype in ['stepfilled', 'bar']:
                    err_color = colors.to_rgba(vis_object[i][0].get_facecolor())
                elif self.histtype == 'step':
                    err_color = colors.to_rgba(vis_object[i][0].get_edgecolor())

                hls_tmp = colorsys.rgb_to_hls(*err_color[:-1])
                err_color = list(colorsys.hls_to_rgb(hls_tmp[0], hls_tmp[1]*0.7, hls_tmp[2])) + \
                    [err_color[-1]]
            elif self.err_dict['err_color'] == 'auto' and self.stacked:
                err_color = next(self.ax._get_lines.prop_cycler)['color']
            else:
                err_color = self.err_dict['err_color']

            if self.histtype == 'marker':
                if self.err_dict['err_x']:
                    xerr = self.widths*0.5
                else:
                    xerr = None
                _, caps, _ = self.ax.errorbar(self.bin_centers, bin_height[i], linestyle='',
                                              marker='', yerr=bin_err[i], xerr=xerr,
                                              linewidth=2, color=err_color)
            else:
                if self.err_dict['err_style'] == 'line':
                    self.ax.errorbar(self.bin_centers, bin_height[i], linestyle='', marker='',
                                     yerr=bin_err[i], linewidth=2, color=err_color)

                elif self.err_dict['err_style'] == 'band':
                    if self.err_dict['err_type'] == 'poisson':
                        fill_between_steps(self.ax, self.bin_edges, bin_height[i]+bin_err[i][1],
                                           bin_height[i]-bin_err[i][0], step_where='pre',
                                           linewidth=0, color=err_color,
                                           alpha=self.hist_dict['alpha']*0.8, zorder=10)
                    else:
                        fill_between_steps(self.ax, self.bin_edges, bin_height[i]+bin_err[i],
                                           bin_height[i]-bin_err[i], step_where='pre', linewidth=0,
                                           color=err_color, alpha=self.hist_dict['alpha']*0.8,
                                           zorder=10)

                    if self.stacked:
                        poly_patch = self.vis_object[-1][0].get_xy()
                        self.ax.add_patch(Polygon(poly_patch[:(len(poly_patch)+1)//2], closed=False,
                                          facecolor='none', edgecolor='k', linewidth=1, alpha=0.5,
                                                  zorder=0))

    def redraw(self):
        with np.errstate(divide='ignore', invalid='ignore'):
            self.bc_scales = np.divide(self.bin_content, self.bin_content_orig)
        self.bc_scales[-1] = np.nan_to_num(self.bc_scales[-1])
        self.do_redraw = False
        if self.n_data_sets == 1:
            bin_content = [self.bin_content]
            if self.histtype != 'marker':
                vis_object = [self.vis_object]
        else:
            bin_content = self.bin_content
            vis_object = self.vis_object

        for n in range(self.n_data_sets):
            if self.stacked:
                xy = vis_object[n][0].get_xy()
                j = 0
                for bcs in self.bc_scales[-1]:
                    xy[j+1, 1] *= bcs
                    xy[j+2, 1] *= bcs
                    j += 2
                if self.histtype == 'step':
                    xy[0, 1] *= self.bc_scales[-1][0]
                    xy[-1, 1] *= self.bc_scales[-1][-1]
                elif self.histtype in ['bar', 'stepfilled']:
                    for bcs in self.bc_scales[-1][::-1]:
                        xy[j+1, 1] *= bcs
                        xy[j+2, 1] *= bcs
                        j += 2
                    xy[0, 1] = xy[-1, 1]
                plt.setp(vis_object[n][0], 'xy', xy)

            else:
                if self.histtype == 'bar':
                    for i, bc in enumerate(bin_content[n]):
                        plt.setp(vis_object[n][i], 'height', bc)
                elif self.histtype == 'stepfilled' or self.histtype == 'step':
                    xy = vis_object[n][0].get_xy()
                    j = 0
                    for bc in bin_content[n]:
                        xy[j+1, 1] = bc
                        xy[j+2, 1] = bc
                        j += 2
                    plt.setp(vis_object[n][0], 'xy', xy)

        self.ax.relim()
        self.ax.autoscale_view(False, False, True)

    def set_log_vals(self):
        self.ax.set_yscale('log', nonposy='clip')
        logbase = self.ax.yaxis._scale.base
        ymin = np.min(self.bin_content[np.nonzero(self.bin_content)])
        try:
            self.ax.set_ylim(bottom=min(self.ax.get_ylim()[0], ymin/logbase))
        except TypeError:
            self.ax.set_ylim(ymin=min(self.ax.get_ylim()[0], ymin/logbase))


def poisson_error(bin_content, suppress_zero=False):
    '''Returns a high and low 1-sigma error bar for an input bin value, as defined in:
    https://www-cdf.fnal.gov/physics/statistics/notes/pois_eb.txt
    If bin_content > 9, returns the sqrt(bin_content)'''
    error_dict = {
        0: (0.000000, 1.000000),
        1: (0.381966, 2.618034),
        2: (1.000000, 4.000000),
        3: (1.697224, 5.302776),
        4: (2.438447, 6.561553),
        5: (3.208712, 7.791288),
        6: (4.000000, 9.000000),
        7: (4.807418, 10.192582),
        8: (5.627719, 11.372281),
        9: (6.458619, 12.541381)}

    if suppress_zero and bin_content == 0:
        return (0, 0)
    elif bin_content in error_dict:
        return error_dict[bin_content]
    else:
        return (np.sqrt(bin_content), np.sqrt(bin_content))



#!/usr/bin/env python
# Licensed under a 3-clause BSD style license, see LICENSE.
import codecs
import os.path
import re


def read(*parts):
    """ Read files """
    file_path = os.path.join(os.path.dirname(__file__), *parts)
    with codecs.open(file_path, 'r') as fobj:
        content = fobj.read()
    return content


def find_version(*parts):
    """ Find version string """
    version_file = read(*parts)
    version_match = re.search(r'^__version__ = [\'"]([^\'"]*)[\'"]',
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError('Unable to find version string.')
